@@ -102,6 +102,7 @@ class AtomApplication extends EventEmitter {
   constructor (options) {
     super()
     this.quitting = false
+    this.quittingForUpdate = false
     this.getAllWindows = this.getAllWindows.bind(this)
     this.getLastFocusedWindow = this.getLastFocusedWindow.bind(this)
     this.resourcePath = options.resourcePath
@@ -198,6 +199,8 @@ class AtomApplication extends EventEmitter {
                (options.urlsToOpen && options.urlsToOpen.length > 0)) {
       optionsForWindowsToOpen.push(options)
       shouldReopenPreviousWindows = this.config.get('core.restorePreviousWindowsOnStart') === 'always'
+    } else if (options.newWindow) {
+      shouldReopenPreviousWindows = false
     } else {
       shouldReopenPreviousWindows = this.config.get('core.restorePreviousWindowsOnStart') !== 'no'
     }
@@ -217,9 +220,9 @@ class AtomApplication extends EventEmitter {
 
   openWithOptions (options) {
     const {
-      initialPaths,
       pathsToOpen,
       executedFrom,
+      foldersToOpen,
       urlsToOpen,
       benchmark,
       benchmarkTest,
@@ -259,10 +262,10 @@ class AtomApplication extends EventEmitter {
         timeout,
         env
       })
-    } else if (pathsToOpen.length > 0) {
+    } else if ((pathsToOpen && pathsToOpen.length > 0) || (foldersToOpen && foldersToOpen.length > 0)) {
       return this.openPaths({
-        initialPaths,
         pathsToOpen,
+        foldersToOpen,
         executedFrom,
         pidToKillWhenClosed,
         newWindow,
@@ -278,7 +281,6 @@ class AtomApplication extends EventEmitter {
     } else {
       // Always open a editor window if this is the first instance of Atom.
       return this.openPath({
-        initialPaths,
         pidToKillWhenClosed,
         newWindow,
         devMode,
@@ -373,13 +375,17 @@ class AtomApplication extends EventEmitter {
 
   // Registers basic application commands, non-idempotent.
   handleEvents () {
-    const getLoadSettings = () => {
+    const getLoadSettings = includeWindow => {
       const window = this.focusedWindow()
-      return {devMode: window && window.devMode, safeMode: window && window.safeMode}
+      return {
+        devMode: window && window.devMode,
+        safeMode: window && window.safeMode,
+        window: includeWindow && window
+      }
     }
 
     this.on('application:quit', () => app.quit())
-    this.on('application:new-window', () => this.openPath(getLoadSettings()))
+    this.on('application:new-window', () => this.openPath(getLoadSettings(false)))
     this.on('application:new-file', () => (this.focusedWindow() || this).openPath())
     this.on('application:open-dev', () => this.promptForPathToOpen('all', {devMode: true}))
     this.on('application:open-safe', () => this.promptForPathToOpen('all', {safeMode: true}))
@@ -397,15 +403,20 @@ class AtomApplication extends EventEmitter {
 
     this.on('application:install-update', () => {
       this.quitting = true
+      this.quittingForUpdate = true
       this.autoUpdateManager.install()
     })
 
     this.on('application:check-for-update', () => this.autoUpdateManager.check())
 
     if (process.platform === 'darwin') {
-      this.on('application:open', () => this.promptForPathToOpen('all', getLoadSettings(), getDefaultPath()))
-      this.on('application:open-file', () => this.promptForPathToOpen('file', getLoadSettings(), getDefaultPath()))
-      this.on('application:open-folder', () => this.promptForPathToOpen('folder', getLoadSettings(), getDefaultPath()))
+      this.on('application:reopen-project', ({ paths }) => {
+        this.openPaths({ pathsToOpen: paths })
+      })
+
+      this.on('application:open', () => this.promptForPathToOpen('all', getLoadSettings(true), getDefaultPath()))
+      this.on('application:open-file', () => this.promptForPathToOpen('file', getLoadSettings(true), getDefaultPath()))
+      this.on('application:open-folder', () => this.promptForPathToOpen('folder', getLoadSettings(true), getDefaultPath()))
       this.on('application:bring-all-windows-to-front', () => Menu.sendActionToFirstResponder('arrangeInFront:'))
       this.on('application:hide', () => Menu.sendActionToFirstResponder('hide:'))
       this.on('application:hide-other-applications', () => Menu.sendActionToFirstResponder('hideOtherApplications:'))
@@ -556,11 +567,11 @@ class AtomApplication extends EventEmitter {
     this.disposable.add(ipcHelpers.on(ipcMain, 'open-command', (event, command, defaultPath) => {
       switch (command) {
         case 'application:open':
-          return this.promptForPathToOpen('all', getLoadSettings(), defaultPath)
+          return this.promptForPathToOpen('all', getLoadSettings(true), defaultPath)
         case 'application:open-file':
-          return this.promptForPathToOpen('file', getLoadSettings(), defaultPath)
+          return this.promptForPathToOpen('file', getLoadSettings(true), defaultPath)
         case 'application:open-folder':
-          return this.promptForPathToOpen('folder', getLoadSettings(), defaultPath)
+          return this.promptForPathToOpen('folder', getLoadSettings(true), defaultPath)
         default:
           return console.log(`Invalid open-command received: ${command}`)
       }
@@ -798,7 +809,6 @@ class AtomApplication extends EventEmitter {
   //   :window - {AtomWindow} to open file paths in.
   //   :addToLastWindow - Boolean of whether this should be opened in last focused window.
   openPath ({
-    initialPaths,
     pathToOpen,
     pidToKillWhenClosed,
     newWindow,
@@ -811,7 +821,6 @@ class AtomApplication extends EventEmitter {
     env
   } = {}) {
     return this.openPaths({
-      initialPaths,
       pathsToOpen: [pathToOpen],
       pidToKillWhenClosed,
       newWindow,
@@ -829,6 +838,7 @@ class AtomApplication extends EventEmitter {
   //
   // options -
   //   :pathsToOpen - The array of file paths to open
+  //   :foldersToOpen - An array of additional paths to open that must be existing directories
   //   :pidToKillWhenClosed - The integer of the pid to kill
   //   :newWindow - Boolean of whether this should be opened in a new window.
   //   :devMode - Boolean to control the opened window's dev mode.
@@ -837,8 +847,8 @@ class AtomApplication extends EventEmitter {
   //   :window - {AtomWindow} to open file paths in.
   //   :addToLastWindow - Boolean of whether this should be opened in last focused window.
   openPaths ({
-    initialPaths,
     pathsToOpen,
+    foldersToOpen,
     executedFrom,
     pidToKillWhenClosed,
     newWindow,
@@ -851,33 +861,57 @@ class AtomApplication extends EventEmitter {
     addToLastWindow,
     env
   } = {}) {
-    if (!pathsToOpen || pathsToOpen.length === 0) return
     if (!env) env = process.env
+    if (!pathsToOpen) pathsToOpen = []
+    if (!foldersToOpen) foldersToOpen = []
+
     devMode = Boolean(devMode)
     safeMode = Boolean(safeMode)
     clearWindowState = Boolean(clearWindowState)
 
-    const locationsToOpen = []
-    for (let i = 0; i < pathsToOpen.length; i++) {
-      const location = this.parsePathToOpen(pathsToOpen[i], executedFrom, addToLastWindow)
-      location.forceAddToWindow = addToLastWindow
-      location.hasWaitSession = pidToKillWhenClosed != null
-      locationsToOpen.push(location)
-      pathsToOpen[i] = location.pathToOpen
+    const locationsToOpen = pathsToOpen.map(pathToOpen => {
+      return this.parsePathToOpen(pathToOpen, executedFrom, {
+        forceAddToWindow: addToLastWindow,
+        hasWaitSession: pidToKillWhenClosed != null
+      })
+    })
+
+    for (const folderToOpen of foldersToOpen) {
+      locationsToOpen.push({
+        pathToOpen: folderToOpen,
+        initialLine: null,
+        initialColumn: null,
+        mustBeDirectory: true,
+        forceAddToWindow: addToLastWindow,
+        hasWaitSession: pidToKillWhenClosed != null
+      })
     }
 
+    if (locationsToOpen.length === 0) {
+      return
+    }
+
+    const normalizedPathsToOpen = locationsToOpen.map(location => location.pathToOpen).filter(Boolean)
+
     let existingWindow
+
+    // Explicitly provided AtomWindow has precedence unless a new window is forced.
     if (!newWindow) {
-      existingWindow = this.windowForPaths(pathsToOpen, devMode)
-      if (!existingWindow) {
-        let lastWindow = window || this.getLastFocusedWindow()
-        if (lastWindow && lastWindow.devMode === devMode) {
-          if (addToLastWindow || (
-              locationsToOpen.every(({stat}) => stat && stat.isFile()) ||
-              (locationsToOpen.some(({stat}) => stat && stat.isDirectory()) && !lastWindow.hasProjectPath()))) {
-            existingWindow = lastWindow
-          }
-        }
+      existingWindow = window
+    }
+
+    // If no window is specified, a new window is not forced, and at least one path is provided, locate
+    // an existing window that contains all paths.
+    if (!existingWindow && !newWindow && normalizedPathsToOpen.length > 0) {
+      existingWindow = this.windowForPaths(normalizedPathsToOpen, devMode)
+    }
+
+    // No window specified, new window not forced, no existing window found, and addition to the last window
+    // requested. Find the last focused window.
+    if (!existingWindow && !newWindow && addToLastWindow) {
+      let lastWindow = window || this.getLastFocusedWindow()
+      if (lastWindow && lastWindow.devMode === devMode) {
+        existingWindow = lastWindow
       }
     }
 
@@ -909,7 +943,6 @@ class AtomApplication extends EventEmitter {
       if (!windowDimensions) windowDimensions = this.getDimensionsForNewWindow()
 
       openedWindow = new AtomWindow(this, this.fileRecoveryService, {
-        initialPaths,
         locationsToOpen,
         windowInitializationScript,
         resourcePath,
@@ -930,7 +963,7 @@ class AtomApplication extends EventEmitter {
       }
       this.waitSessionsByWindow.get(openedWindow).push({
         pid: pidToKillWhenClosed,
-        remainingPaths: new Set(pathsToOpen)
+        remainingPaths: new Set(normalizedPathsToOpen)
       })
     }
 
@@ -998,8 +1031,7 @@ class AtomApplication extends EventEmitter {
     const states = await this.storageFolder.load('application.json')
     if (states) {
       return states.map(state => ({
-        initialPaths: state.initialPaths,
-        pathsToOpen: state.initialPaths.filter(p => fs.isDirectorySync(p)),
+        foldersToOpen: state.initialPaths,
         urlsToOpen: [],
         devMode: this.devMode,
         safeMode: this.safeMode
@@ -1278,7 +1310,7 @@ class AtomApplication extends EventEmitter {
     }
   }
 
-  parsePathToOpen (pathToOpen, executedFrom = '') {
+  parsePathToOpen (pathToOpen, executedFrom, extra) {
     let initialColumn, initialLine
     if (!pathToOpen) {
       return {pathToOpen}
@@ -1300,10 +1332,9 @@ class AtomApplication extends EventEmitter {
     }
 
     const normalizedPath = path.normalize(path.resolve(executedFrom, fs.normalize(pathToOpen)))
-    const stat = fs.statSyncNoException(normalizedPath)
-    if (stat || !url.parse(pathToOpen).protocol) pathToOpen = normalizedPath
+    if (!url.parse(pathToOpen).protocol) pathToOpen = normalizedPath
 
-    return {pathToOpen, stat, initialLine, initialColumn}
+    return Object.assign({pathToOpen, initialLine, initialColumn}, extra)
   }
 
   // Opens a native dialog to prompt the user for a path.
@@ -1317,14 +1348,32 @@ class AtomApplication extends EventEmitter {
   //              should be in dev mode or not.
   //   :safeMode - A Boolean which controls whether any newly opened windows
   //               should be in safe mode or not.
-  //   :window - An {AtomWindow} to use for opening a selected file path.
+  //   :window - An {AtomWindow} to use for opening selected file paths as long as
+  //             all are files.
   //   :path - An optional String which controls the default path to which the
   //           file dialog opens.
   promptForPathToOpen (type, {devMode, safeMode, window}, path = null) {
     return this.promptForPath(
       type,
-      pathsToOpen => {
-        return this.openPaths({pathsToOpen, devMode, safeMode, window})
+      async pathsToOpen => {
+        let targetWindow
+
+        // Open in :window as long as no chosen paths are folders. If any chosen path is a folder, open in a
+        // new window instead.
+        if (type === 'folder') {
+          targetWindow = null
+        } else if (type === 'file') {
+          targetWindow = window
+        } else if (type === 'all') {
+          const areDirectories = await Promise.all(
+            pathsToOpen.map(pathToOpen => new Promise(resolve => fs.isDirectory(pathToOpen, resolve)))
+          )
+          if (!areDirectories.some(Boolean)) {
+            targetWindow = window
+          }
+        }
+
+        return this.openPaths({pathsToOpen, devMode, safeMode, window: targetWindow})
       },
       path
     )
